#
#  Copyright 2021. Logical Clocks AB
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from __future__ import annotations

import copy
import json
import logging
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import humps
from hopsworks_common import client
from hopsworks_common.client.exceptions import FeatureStoreException
from hopsworks_common.constants import FEATURES
from hsfs import util
from hsfs.core import transformation_function_engine
from hsfs.core.feature_descriptive_statistics import FeatureDescriptiveStatistics
from hsfs.decorators import typechecked
from hsfs.hopsworks_udf import HopsworksUdf
from hsfs.transformation_statistics import TransformationStatistics
from packaging.version import Version


_logger = logging.getLogger(__name__)


class TransformationType(Enum):
    """
    Class that store the possible types of transformation functions.
    """

    MODEL_DEPENDENT = "model_dependent"
    ON_DEMAND = "on_demand"
    UNDEFINED = "undefined"  # This type is used when the UDF created is not attached to a feature view / feature group. Hence the transformation function is neither model dependent nor on-demand.


@typechecked
class TransformationFunction:
    NOT_FOUND_ERROR_CODE = 270160
    """
    DTO class for transformation functions.

    # Arguments
        featurestore_id : `int`. Id of the feature store in which the transformation function is saved.
        hopsworks_udf : `HopsworksUDF`. The meta data object for UDF in Hopsworks, which can be created using the `@udf` decorator.
        version : `int`. The version of the transformation function.
        id : `int`. The id of the transformation function in the feature store.
        transformation_type : `UDFType`. The type of the transformation function. Can be "on-demand" or "model-dependent"
    """

    def __init__(
        self,
        featurestore_id: int,
        hopsworks_udf: HopsworksUdf,
        version: Optional[int] = None,
        id: Optional[int] = None,
        transformation_type: Optional[TransformationType] = None,
        type=None,
        items=None,
        count=None,
        href=None,
        **kwargs,
    ):
        self._id: int = id
        self._featurestore_id: int = featurestore_id
        self._version: int = version

        self._transformation_function_engine = (
            transformation_function_engine.TransformationFunctionEngine(
                self._featurestore_id
            )
        )
        if not isinstance(hopsworks_udf, HopsworksUdf):
            raise FeatureStoreException(
                "Please use the hopsworks_udf decorator when defining transformation functions."
            )

        self.__hopsworks_udf: HopsworksUdf = hopsworks_udf
        TransformationFunction._validate_transformation_type(
            transformation_type=transformation_type, hopsworks_udf=hopsworks_udf
        )

        # Setting transformation type as unknown when the transformation function is not attached to a feature view / feature group.
        # This happens for example when the transformation function is fetched from the backend.
        self.transformation_type = (
            transformation_type if transformation_type else TransformationType.UNDEFINED
        )

        if self.__hopsworks_udf._generate_output_col_name:
            # Reset output column names so that they would be regenerated.
            # Handles the use case in which the same UDF is used to define both on-demand and model dependent transformations.
            self.__hopsworks_udf._output_column_names = []

    def save(self) -> None:
        """Save a transformation function into the backend.

        !!! example
            ```python
            # import hopsworks udf decorator
            from hopworks import udf

            # define function
            @udf(int)
            def plus_one(value):
                return value + 1

            # create transformation function
            plus_one_meta = fs.create_transformation_function(
                    transformation_function=plus_one,
                    version=1
                )

            # persist transformation function in backend
            plus_one_meta.save()
            ```
        """
        self._transformation_function_engine.save(self)

    def delete(self) -> None:
        """Delete transformation function from backend.

        !!! example
            ```python
            # import hopsworks udf decorator
            from hopworks import udf

            # define function
            @udf(int)
            def plus_one(value):
                return value + 1

            # create transformation function
            plus_one_meta = fs.create_transformation_function(
                    transformation_function=plus_one,
                    version=1
                )
            # persist transformation function in backend
            plus_one_meta.save()

            # retrieve transformation function
            plus_one_fn = fs.get_transformation_function(name="plus_one")

            # delete transformation function from backend
            plus_one_fn.delete()
            ```
        """
        self._transformation_function_engine.delete(self)

    def __call__(self, *features: List[str]) -> TransformationFunction:
        """
        Update the feature to be using in the transformation function

        # Arguments
            features: `List[str]`. Name of features to be passed to the User Defined function
        # Returns
            `HopsworksUdf`: Meta data class for the user defined function.
        # Raises
            `hopsworks.client.exceptions.FeatureStoreException`: If the provided number of features do not match the number of arguments in the defined UDF or if the provided feature names are not strings.
        """
        # Deep copy so that the same transformation function can be used to create multiple new transformation function with different features.
        transformation = copy.deepcopy(self)
        transformation.__hopsworks_udf = transformation.__hopsworks_udf(*features)

        return transformation

    @classmethod
    def from_response_json(
        cls, json_dict: Dict[str, Any]
    ) -> Union[TransformationFunction, List[TransformationFunction]]:
        """
        Function that constructs the class object from its json serialization.

        # Arguments
            json_dict: `Dict[str, Any]`. Json serialized dictionary for the class.
        # Returns
            `TransformationFunction`: Json deserialized class object.
        """
        json_decamelized = humps.decamelize(json_dict)

        if "count" in json_decamelized:
            if json_decamelized["count"] == 0:
                return []
            for tffn_dto in json_decamelized["items"]:
                if tffn_dto.get("hopsworks_udf", False):
                    tffn_dto["hopsworks_udf"] = HopsworksUdf.from_response_json(
                        tffn_dto["hopsworks_udf"]
                    )
            if json_decamelized["count"] == 1:
                return cls(**json_decamelized["items"][0])
            else:
                return [cls(**tffn_dto) for tffn_dto in json_decamelized["items"]]
        else:
            if json_decamelized.get("hopsworks_udf", False):
                json_decamelized["hopsworks_udf"] = HopsworksUdf.from_response_json(
                    json_decamelized["hopsworks_udf"]
                )
            return cls(**json_decamelized)

    def update_from_response_json(
        self, json_dict: Dict[str, Any]
    ) -> TransformationFunction:
        """
        Function that updates the class object from its json serialization.

        # Arguments
            json_dict: `Dict[str, Any]`. Json serialized dictionary for the class.
        # Returns
            `TransformationFunction`: Json deserialized class object.
        """
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(**json_decamelized)
        return self

    def json(self) -> str:
        """
        Convert class into its json serialized form.

        # Returns
            `str`: Json serialized object.
        """
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert class into a dictionary.

        # Returns
            `Dict`: Dictionary that contains all data required to json serialize the object.
        """
        backend_version = client.get_connection().backend_version

        return {
            "id": self._id,
            "version": self._version,
            "featurestoreId": self._featurestore_id,
            "hopsworksUdf": self.hopsworks_udf.to_dict(),
            **(
                {"transformationType": self.transformation_type.value}
                if Version(backend_version) > Version("4.1.6")
                else {}
            ),  # This check is added for backward compatibility with older versions of Hopsworks. The "transformationType" field was added for equality checking of transformation functions and versions below 4.1.6 do not support unknown fields in the backend.
        }

    def alias(self, *args: str):
        """
        Set the names of the transformed features output by the transformation function.
        """
        self.__hopsworks_udf.alias(*args)

        return self

    def _get_output_column_names(self) -> str:
        """
        Function that generates feature names for the transformed features

        # Returns
            `List[str]`: List of feature names for the transformed columns
        """
        # If function name matches the name of an input feature and the transformation function only returns one output feature then
        # then the transformed output feature would have the same name as the input feature. i.e the input feature will get overwritten.
        if (
            len(self.__hopsworks_udf.return_types) == 1
            and any(
                [
                    self.__hopsworks_udf.function_name
                    == transformation_feature.feature_name
                    for transformation_feature in self.__hopsworks_udf._transformation_features
                ]
            )
            and (
                not self.__hopsworks_udf.dropped_features
                or self.__hopsworks_udf.function_name
                not in self.__hopsworks_udf.dropped_features
            )
        ):
            output_col_names = [self.__hopsworks_udf.function_name]
        else:
            if self.transformation_type == TransformationType.MODEL_DEPENDENT:
                _BASE_COLUMN_NAME = f'{self.__hopsworks_udf.function_name}_{"_".join(self.__hopsworks_udf.transformation_features)}_'
            elif self.transformation_type == TransformationType.ON_DEMAND:
                _BASE_COLUMN_NAME = (
                    self.__hopsworks_udf.function_name
                    if len(self.__hopsworks_udf.return_types) == 1
                    else f"{self.__hopsworks_udf.function_name}_"
                )

            output_col_names = (
                [
                    f"{_BASE_COLUMN_NAME}{i}"
                    for i in range(len(self.__hopsworks_udf.return_types))
                ]
                if len(self.__hopsworks_udf.return_types) > 1
                else [_BASE_COLUMN_NAME]
            )

        if any(
            len(output_col_name) > FEATURES.MAX_LENGTH_NAME
            for output_col_name in output_col_names
        ):
            _logger.warning(
                f"The default output feature names generated by the transformation function {repr(self.__hopsworks_udf)} exceed the maximum allowed length of {FEATURES.MAX_LENGTH_NAME} characters. Default names have been truncated to fit within the size limit. To avoid this, consider using the alias function to explicitly specify output column names."
            )
            if len(output_col_names) > 1:
                # Slicing the output column names
                for index, output_col_name in enumerate(output_col_names):
                    output_col_names[index] = (
                        f"{output_col_name[:FEATURES.MAX_LENGTH_NAME-(len(output_col_names) + 1)]}_{str(index)}"
                        if len(output_col_name) > FEATURES.MAX_LENGTH_NAME
                        else output_col_name
                    )
            else:
                output_col_names = [output_col_names[0][: FEATURES.MAX_LENGTH_NAME]]

        return output_col_names

    @staticmethod
    def _validate_transformation_type(
        transformation_type: TransformationType, hopsworks_udf: HopsworksUdf
    ):
        """
        Function that returns validates if the defined transformation function can be used for the specified UDF type.

        # Raises
            `hopsworks.client.exceptions.FeatureStoreException` : If the UDF Type is None or if statistics or multiple columns has been output by a on-demand transformation function
        """

        if transformation_type == TransformationType.ON_DEMAND:
            if hopsworks_udf.statistics_required:
                raise FeatureStoreException(
                    "On-Demand Transformation functions cannot use statistics, please remove statistics parameters from the functions"
                )

    @property
    def id(self) -> id:
        """Transformation function id."""
        return self._id

    @id.setter
    def id(self, id: int) -> None:
        self._id = id

    @property
    def version(self) -> int:
        """Version of the transformation function."""
        return self._version

    @version.setter
    def version(self, version: int) -> None:
        self._version = version

    @property
    def hopsworks_udf(self) -> HopsworksUdf:
        """Meta data class for the user defined transformation function."""
        # Make sure that the output column names for a model-dependent or on-demand transformation function, when accessed externally from the class.
        if (
            self.transformation_type
            and self.transformation_type != TransformationType.UNDEFINED
            and not self.__hopsworks_udf.output_column_names
        ):
            self.__hopsworks_udf.output_column_names = self._get_output_column_names()
        return self.__hopsworks_udf

    @property
    def transformation_type(self) -> TransformationType:
        """Type of the Transformation : Can be \"model dependent\" or \"on-demand\" """
        return self._transformation_type

    @transformation_type.setter
    def transformation_type(self, transformation_type) -> None:
        self._transformation_type = transformation_type

    @property
    def transformation_statistics(
        self,
    ) -> Optional[TransformationStatistics]:
        """Feature statistics required for the defined UDF"""
        return self.__hopsworks_udf.transformation_statistics

    @transformation_statistics.setter
    def transformation_statistics(
        self, statistics: List[FeatureDescriptiveStatistics]
    ) -> None:
        self.__hopsworks_udf.transformation_statistics = statistics
        # Generate output column names for one-hot encoder after transformation statistics is set.
        # This is done because the number of output columns for one-hot encoding dependents on number of unique values in training dataset statistics.
        if self.__hopsworks_udf.function_name == "one_hot_encoder":
            self.__hopsworks_udf.output_column_names = self._get_output_column_names()

    @property
    def output_column_names(self) -> List[str]:
        """Names of the output columns generated by the transformation functions"""
        if (
            self.__hopsworks_udf.function_name == "one_hot_encoder"
            and len(self.__hopsworks_udf.output_column_names)
            != len(self.__hopsworks_udf.return_types)
        ) or not self.__hopsworks_udf.output_column_names:
            self.__hopsworks_udf.output_column_names = self._get_output_column_names()
        return self.__hopsworks_udf.output_column_names

    def __repr__(self):
        if self.transformation_type == TransformationType.MODEL_DEPENDENT:
            return f"Model-Dependent Transformation Function : {repr(self.__hopsworks_udf)}"
        elif self.transformation_type == TransformationType.ON_DEMAND:
            return f"On-Demand Transformation Function : {repr(self.__hopsworks_udf)}"
        else:
            return f"Transformation Function : {repr(self.__hopsworks_udf)}"

    def __eq__(self, other):
        return self.to_dict() == other.to_dict()



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import ast
import datetime
import decimal
from typing import TYPE_CHECKING, Literal, Union

import pytz
from hopsworks_common.core.constants import (
    HAS_PANDAS,
    HAS_POLARS,
    HAS_PYARROW,
)
from hopsworks_common.decorators import uses_polars


if TYPE_CHECKING:
    import numpy as np
    import pandas as pd
    import polars as pl

if HAS_PYARROW:
    import pyarrow as pa

    # Decimal types are currently not supported
    _INT_TYPES = [pa.uint8(), pa.uint16(), pa.int8(), pa.int16(), pa.int32()]
    _BIG_INT_TYPES = [pa.uint32(), pa.int64()]
    _FLOAT_TYPES = [pa.float16(), pa.float32()]
    _DOUBLE_TYPES = [pa.float64()]
    _TIMESTAMP_UNIT = ["ns", "us", "ms", "s"]
    _BOOLEAN_TYPES = [pa.bool_()]
    _STRING_TYPES = [pa.string(), pa.large_string()]
    _DATE_TYPES = [pa.date32(), pa.date64()]
    _BINARY_TYPES = [pa.binary(), pa.large_binary()]

    PYARROW_HOPSWORKS_DTYPE_MAPPING = {
        **dict.fromkeys(_INT_TYPES, "int"),
        **dict.fromkeys(_BIG_INT_TYPES, "bigint"),
        **dict.fromkeys(_FLOAT_TYPES, "float"),
        **dict.fromkeys(_DOUBLE_TYPES, "double"),
        **dict.fromkeys(
            [
                *[pa.timestamp(unit) for unit in _TIMESTAMP_UNIT],
                *[
                    pa.timestamp(unit, tz=tz)
                    for unit in _TIMESTAMP_UNIT
                    for tz in pytz.all_timezones
                ],
            ],
            "timestamp",
        ),
        **dict.fromkeys(_BOOLEAN_TYPES, "boolean"),
        **dict.fromkeys(
            [
                *_STRING_TYPES,
                # Category type in pandas stored as dictinoary in pyarrow
                *[
                    pa.dictionary(
                        value_type=value_type, index_type=index_type, ordered=ordered
                    )
                    for value_type in _STRING_TYPES
                    for index_type in _INT_TYPES + _BIG_INT_TYPES
                    for ordered in [True, False]
                ],
            ],
            "string",
        ),
        **dict.fromkeys(_DATE_TYPES, "date"),
        **dict.fromkeys(_BINARY_TYPES, "binary"),
    }
else:
    PYARROW_HOPSWORKS_DTYPE_MAPPING = {}

# python cast column to offline type
if HAS_POLARS:
    import polars as pl

    polars_offline_dtype_mapping = {
        "bigint": pl.Int64,
        "int": pl.Int32,
        "smallint": pl.Int16,
        "tinyint": pl.Int8,
        "float": pl.Float32,
        "double": pl.Float64,
    }

    _polars_online_dtype_mapping = {
        "bigint": pl.Int64,
        "int": pl.Int32,
        "smallint": pl.Int16,
        "tinyint": pl.Int8,
        "float": pl.Float32,
        "double": pl.Float64,
    }

if HAS_PANDAS:
    import numpy as np
    import pandas as pd

    pandas_offline_dtype_mapping = {
        "bigint": pd.Int64Dtype(),
        "int": pd.Int32Dtype(),
        "smallint": pd.Int16Dtype(),
        "tinyint": pd.Int8Dtype(),
        "float": np.dtype("float32"),
        "double": np.dtype("float64"),
    }

    pandas_online_dtype_mapping = {
        "bigint": pd.Int64Dtype(),
        "int": pd.Int32Dtype(),
        "smallint": pd.Int16Dtype(),
        "tinyint": pd.Int8Dtype(),
        "float": np.dtype("float32"),
        "double": np.dtype("float64"),
    }


def convert_pandas_dtype_to_offline_type(arrow_type: str) -> str:
    # This is a simple type conversion between pandas dtypes and pyspark (hive) types,
    # using pyarrow types obatined from pandas dataframe to convert pandas typed fields,
    # A recurisive function  "convert_pandas_object_type_to_offline_type" is used to convert complex types like lists and structures
    # "_onvert_simple_pandas_dtype_to_offline_type" is used to convert simple types
    # In the backend, the types specified here will also be used for mapping to Avro types.
    if (
        pa.types.is_list(arrow_type)
        or pa.types.is_large_list(arrow_type)
        or pa.types.is_struct(arrow_type)
    ):
        return convert_pandas_object_type_to_offline_type(arrow_type)

    return convert_simple_pandas_dtype_to_offline_type(arrow_type)


def convert_pandas_object_type_to_offline_type(arrow_type: str) -> str:
    if pa.types.is_list(arrow_type) or pa.types.is_large_list(arrow_type):
        # figure out sub type
        sub_arrow_type = arrow_type.value_type
        subtype = convert_pandas_dtype_to_offline_type(sub_arrow_type)
        return "array<{}>".format(subtype)
    if pa.types.is_struct(arrow_type):
        struct_schema = {}
        for index in range(arrow_type.num_fields):
            struct_schema[arrow_type.field(index).name] = (
                convert_pandas_dtype_to_offline_type(arrow_type.field(index).type)
            )
        return (
            "struct<"
            + ",".join([f"{key}:{value}" for key, value in struct_schema.items()])
            + ">"
        )

    raise ValueError(f"dtype 'O' (arrow_type '{str(arrow_type)}') not supported")


def cast_pandas_column_to_offline_type(
    feature_column: pd.Series, offline_type: str
) -> pd.Series:
    offline_type = offline_type.lower()
    if offline_type == "timestamp":
        return pd.to_datetime(feature_column, utc=True).dt.tz_localize(None)
    elif offline_type == "date":
        return pd.to_datetime(feature_column, utc=True).dt.date
    elif (
        offline_type.startswith("array<")
        or offline_type.startswith("struct<")
        or offline_type == "boolean"
    ):
        return feature_column.apply(
            lambda x: (ast.literal_eval(x) if isinstance(x, str) else x)
            if (x is not None and x != "")
            else None
        )
    elif offline_type == "string":
        return feature_column.apply(lambda x: str(x) if x is not None else None)
    elif offline_type.startswith("decimal"):
        return feature_column.apply(
            lambda x: decimal.Decimal(x) if (x is not None) else None
        )
    else:
        if offline_type in pandas_offline_dtype_mapping:
            return feature_column.astype(pandas_offline_dtype_mapping[offline_type])
        else:
            return feature_column  # handle gracefully, just return the column as-is


@uses_polars
def cast_polars_column_to_offline_type(
    feature_column: pl.Series, offline_type: str
) -> pl.Series:
    offline_type = offline_type.lower()
    if offline_type == "timestamp":
        # convert (if tz!=UTC) to utc, then make timezone unaware
        return feature_column.cast(pl.Datetime(time_zone=None))
    elif offline_type == "date":
        return feature_column.cast(pl.Date)
    elif (
        offline_type.startswith("array<")
        or offline_type.startswith("struct<")
        or offline_type == "boolean"
    ):
        return feature_column.map_elements(
            lambda x: (ast.literal_eval(x) if isinstance(x, str) else x)
            if (x is not None and x != "")
            else None
        )
    elif offline_type == "string":
        return feature_column.map_elements(lambda x: str(x) if x is not None else None)
    elif offline_type.startswith("decimal"):
        return feature_column.map_elements(
            lambda x: decimal.Decimal(x) if (x is not None) else None
        )
    else:
        if offline_type in polars_offline_dtype_mapping:
            return feature_column.cast(polars_offline_dtype_mapping[offline_type])
        else:
            return feature_column  # handle gracefully, just return the column as-is


def cast_column_to_offline_type(
    feature_column: Union[pd.Series, pl.Series], offline_type: str
) -> pd.Series:
    if isinstance(feature_column, pd.Series):
        return cast_pandas_column_to_offline_type(feature_column, offline_type.lower())
    elif HAS_POLARS and isinstance(feature_column, pl.Series):
        return cast_polars_column_to_offline_type(feature_column, offline_type.lower())


def cast_column_to_online_type(
    feature_column: pd.Series, online_type: str
) -> pd.Series:
    online_type = online_type.lower()
    if online_type == "timestamp":
        # convert (if tz!=UTC) to utc, then make timezone unaware
        return pd.to_datetime(feature_column, utc=True).dt.tz_localize(None)
    elif online_type == "date":
        return pd.to_datetime(feature_column, utc=True).dt.date
    elif online_type.startswith("varchar") or online_type == "text":
        return feature_column.apply(lambda x: str(x) if x is not None else None)
    elif online_type == "boolean":
        return feature_column.apply(
            lambda x: (ast.literal_eval(x) if isinstance(x, str) else x)
            if (x is not None and x != "")
            else None
        )
    elif online_type.startswith("decimal"):
        return feature_column.apply(
            lambda x: decimal.Decimal(x) if (x is not None) else None
        )
    else:
        if online_type in pandas_online_dtype_mapping:
            casted_feature = feature_column.astype(
                pandas_online_dtype_mapping[online_type]
            )
            return casted_feature
        else:
            return feature_column  # handle gracefully, just return the column as-is


def convert_simple_pandas_dtype_to_offline_type(arrow_type: str) -> str:
    try:
        return PYARROW_HOPSWORKS_DTYPE_MAPPING[arrow_type]
    except KeyError as err:
        raise ValueError(f"dtype '{arrow_type}' not supported") from err


def translate_legacy_spark_type(
    output_type: str,
) -> Literal[
    "STRING",
    "BINARY",
    "BYTE",
    "SHORT",
    "INT",
    "LONG",
    "FLOAT",
    "DOUBLE",
    "TIMESTAMP",
    "DATE",
    "BOOLEAN",
]:
    if output_type == "StringType()":
        return "STRING"
    elif output_type == "BinaryType()":
        return "BINARY"
    elif output_type == "ByteType()":
        return "BYTE"
    elif output_type == "ShortType()":
        return "SHORT"
    elif output_type == "IntegerType()":
        return "INT"
    elif output_type == "LongType()":
        return "LONG"
    elif output_type == "FloatType()":
        return "FLOAT"
    elif output_type == "DoubleType()":
        return "DOUBLE"
    elif output_type == "TimestampType()":
        return "TIMESTAMP"
    elif output_type == "DateType()":
        return "DATE"
    elif output_type == "BooleanType()":
        return "BOOLEAN"
    else:
        return "STRING"  # handle gracefully, and return STRING type, the default for spark udfs


def convert_spark_type_to_offline_type(spark_type_string: str) -> str:
    if spark_type_string.endswith("Type()"):
        spark_type_string = translate_legacy_spark_type(spark_type_string)
    if spark_type_string == "STRING":
        return "STRING"
    elif spark_type_string == "BINARY":
        return "BINARY"
    elif spark_type_string == "BYTE":
        return "INT"
    elif spark_type_string == "SHORT":
        return "INT"
    elif spark_type_string == "INT":
        return "INT"
    elif spark_type_string == "LONG":
        return "BIGINT"
    elif spark_type_string == "FLOAT":
        return "FLOAT"
    elif spark_type_string == "DOUBLE":
        return "DOUBLE"
    elif spark_type_string == "TIMESTAMP":
        return "TIMESTAMP"
    elif spark_type_string == "DATE":
        return "DATE"
    elif spark_type_string == "BOOLEAN":
        return "BOOLEAN"
    else:
        raise ValueError(
            f"Return type {spark_type_string} not supported for transformation functions."
        )


def infer_spark_type(output_type):
    if not output_type:
        return "STRING"  # STRING is default type for spark udfs

    if isinstance(output_type, str):
        if output_type.endswith("Type()"):
            return translate_legacy_spark_type(output_type)
        output_type = output_type.lower()

    if output_type in (str, "str", "string"):
        return "STRING"
    elif output_type in (bytes, "binary"):
        return "BINARY"
    elif output_type in (np.int8, "int8", "byte", "tinyint"):
        return "BYTE"
    elif output_type in (np.int16, "int16", "short", "smallint"):
        return "SHORT"
    elif output_type in (int, "int", "integer", np.int32):
        return "INT"
    elif output_type in (np.int64, "int64", "long", "bigint"):
        return "LONG"
    elif output_type in (float, "float"):
        return "FLOAT"
    elif output_type in (np.float64, "float64", "double"):
        return "DOUBLE"
    elif output_type in (
        datetime.datetime,
        np.datetime64,
        "datetime",
        "timestamp",
    ):
        return "TIMESTAMP"
    elif output_type in (datetime.date, "date"):
        return "DATE"
    elif output_type in (bool, "boolean", "bool"):
        return "BOOLEAN"
    else:
        raise TypeError("Not supported type %s." % output_type)
